@@ -6,7 +6,7 @@ use crate::Error;
 use futures::{Stream, StreamExt};
 use std::{path::Path, sync::Arc};
 use tokio::fs::{self, File};
-use tokio::io::copy;
+use tokio::io::{copy, AsyncWriteExt};
 
 /// Accepts a stream of future file `parts` and concatenates them into the `dest` file.
 pub async fn concatenator<P>(dest: &mut File, mut parts: P) -> Result<(), Error>
@@ -18,6 +18,8 @@ where
         concatenate(dest, part_path).await?;
     }
 
+    let _ = dest.flush().await;
+
     Ok(())
 }
 
